# -*- coding: utf-8 -*-
# Copyright 2011 Thamini S.à.R.L    This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import time
from datetime import datetime
from dateutil.relativedelta import relativedelta
import logging
import operator

from osv import osv
from osv import fields
import decimal_precision as dp
from tools.misc import DEFAULT_SERVER_DATETIME_FORMAT, DEFAULT_SERVER_DATE_FORMAT
import netsvc
import tools

class config_handled_fields_function(fields.function):
    def __init__(self, fnct, arg=None, fnct_inv=None, fnct_inv_arg=None, type='float', fnct_search=None, obj=None, method=False, store=False, multi=False, **args):
        super(config_handled_fields_function, self).__init__(fnct, arg=arg,
                                                            fnct_inv=fnct_inv,
                                                            fnct_inv_arg=fnct_inv_arg,
                                                            type=type,
                                                            fnct_search=fnct_search,
                                                            obj=obj,
                                                            method=method,
                                                            store=store,
                                                            multi=multi,
                                                            **args)
        if self.store:
            self._classic_read = False
            self._classic_writer = False


class insurance_contract_line__advance(osv.osv):
    _name = 'insurance.contract.line'
insurance_contract_line__advance()

class insurance_expert_report__advance(osv.osv):
    _name = 'insurance.expert.report'
insurance_expert_report__advance()

class insurance_claim__advance(osv.osv):
    _name = 'insurance.claim'
insurance_claim__advance()

class insurance_contract_type_line__advance(osv.osv):
    _name = 'insurance.contract.type.line'
insurance_contract_type_line__advance()

class insurance_contract_type_specific_conditions_attachments__advance(osv.osv):
    _name = 'insurance.contract.type.specific.conditions.attachments'
insurance_contract_type_specific_conditions_attachments__advance()

class insurance_contract_clause(osv.osv):
    _name = 'insurance.contract.clause'
    _columns = {
        'code': fields.char('Code', size=32),
        'name': fields.char('Name', size=64),
    }
insurance_contract_clause()

class insurance_contract_type(osv.osv):
    _name = 'insurance.contract.type'

    _validation_types_selection = [
        ('simple', 'Simple'),
        ('double', 'Double'),
    ]

    _periodicity_types_selection = [
        ('days', 'Days'),
        ('weeks', 'Weeks'),
        ('months', 'Months'),
        ('years', 'Years'),
    ]

    _columns = {
        'name': fields.char('Name', size=32, required=True),
        'code': fields.char('Code', size=16, required=True),
        'internal_description': fields.text('Description'),
        'reinsurer_needed': fields.boolean('Reinsurer Needed',
                                help='If checked, the reinsurer field will be mandatory on the contrat'),
        'reinsurer_id': fields.many2one('res.users', 'Reinsurer',
                                help='If filled, this is the default reinsurer which will be used for this contract type'),
        'validation_type': fields.selection(_validation_types_selection, 'Validation Type', required=True,
                                help='Determine if a second (double) validation is required (normally by the reinsurer)'),
        'forced_start_date': fields.date('Forced Start Date',
                                help='the default beginning date of the contract (note that the system does not take into account the year (only the day and month))'),
        'periodicity_interval': fields.integer('Periodicity Internal'),
        'periodicity_interval_type': fields.selection(_periodicity_types_selection, 'Periodicity Type'),
        'tmpl_line_ids': fields.one2many('insurance.contract.type.line', 'contract_type_id', 'Template Lines',
                                help='specify the default products which will be added to a new contract'),
        'tmpl_clause_ids': fields.many2many('insurance.contract.clause', 'contract_type_id', 'contract_clause_id', 'insurance_contract_type_clause_rel', 'Template Clauses',
                                help='Specify the default clauses which will be added to a new contract'),
        'specific_conditions_attachments': fields.one2many('insurance.contract.type.specific.conditions.attachments', 'type_id',
                                                            'Specific Conditions Attachments',
                                                            help='The following documents will be attachemnt to the printed contract'),
        'deductible_amount': fields.float('Deductible Amount', digits_compute=dp.get_precision('Amount')),
    }

    _defaults = {
        'validation_type': 'simple',
        'periodicity_interval': 1,
    }

    _periodicity_relativedelta = {
        'days': lambda interval: relativedelta(days=interval-1),
        'weeks': lambda interval: relativedelta(days=7*interval-1),
        'months': lambda interval: relativedelta(months=interval, days=-1),
        'years': lambda interval: relativedelta(years=interval, days=-1),
    }

    def get_end_date_str(self, cr, uid, ids, start_date=None, context=None):
        if not start_date:
            start_date = time.strftime(DEFAULT_SERVER_DATE_FORMAT)
        dt_start_date = datetime.strptime(start_date, DEFAULT_SERVER_DATE_FORMAT)
        result = {}
        for contract_type in self.browse(cr, uid, ids, context=context):
            relative_delta_fnct = self._periodicity_relativedelta.get(contract_type.periodicity_interval_type, lambda x: relativedelta(days=0))
            st_end_date = dt_start_date + relative_delta_fnct(contract_type.periodicity_interval)
            result[contract_type.id] = st_end_date.strftime(DEFAULT_SERVER_DATE_FORMAT)
        return result

insurance_contract_type()

class insurance_contract_type_line(osv.osv):
    _name = 'insurance.contract.type.line'

    def _get_price_subtotal(self, cr, uid, ids, fname, args, context=None):
        res = dict.fromkeys(ids, 0.0)
        for l in self.browse(cr, uid, ids, context=context):
            res[l.id] = l.product_qty * l.price_unit
        return res

    def onchange_product_id(self, cr, uid, ids, product_id, product_uom, context=None):
        if context is None:
            context = {}
        property_pool = self.pool.get('ir.property')
        ocv = { 'value': {}}
        if not product_id:
            ocv['value'].update({
                'product_uom': False,
                'account_id': False,
                'tax_ids': False,
            })
            return ocv
        p = self.pool.get('product.product').browse(cr, uid, product_id, context=context)
        ocv['value'].update({
            'price_unit': p.list_price,
            'product_uom': p.uom_id.id or product_uom or False,
            'note': p.description,
            'tax_ids': [ t.id for t in p.taxes_id ],
            'account_id': p.product_tmpl_id.property_account_income.id or p.categ_id.property_account_income_categ.id or False,
        })
        return ocv

    def onchange_product_qty(self, cr, uid, ids, product_id, product_uom, product_qty, price_unit, context=None):
        ocv = {'value': {}}
        # TODO: take care of product_id, product_uom
        ocv['value']['price_subtotal'] = product_qty * price_unit
        return ocv

    _columns = {
        'contract_type_id': fields.many2one('insurance.contract.type', 'Type', required=True, ondelete='cascade'),
        'product_id': fields.many2one('product.product', 'Product', required=True),
        'account_id': fields.many2one('account.account', 'Account', required=True),
        'tax_ids': fields.many2many('account.tax', 'insurance_contract_line_tax_rel', 'contract_line_id', 'tax_id', 'Taxes', domain=[('parent_id','=',False),('type_tax_use','in',['sale','all'])]),
        'product_uom': fields.many2one('product.uom', 'UoM'),
        'product_qty': fields.float('Quantity', required=True),
        'price_unit': fields.float('Unit Price', required=True, digits_compute=dp.get_precision('Account')),
        'price_subtotal': fields.function(_get_price_subtotal, type='float', method=True, string='Untaxed Subtotal', digits_compute=dp.get_precision('Account')),
        'note': fields.text('Note'),
    }

    _defaults = {
        'product_qty': 1.0,
    }
insurance_contract_type_line()

class insurance_contract_type_specific_conditions_attachments(osv.osv):
    _name = 'insurance.contract.type.specific.conditions.attachments'
    _order = 'sequence ASC'
    _columns = {
        'sequence': fields.integer('Sequence', required=True),
        'type_id': fields.many2one('insurance.contract.type', 'Contract Type'),
        'attachment_id': fields.many2one('ir.attachment', 'Attachment', domain="[('res_model','=','insurance.contract.type'),('res_id','=',type_id)]",
                                        ondelete='cascade'),
        'lang_id': fields.many2one('res.lang', 'Language', help='If specified, the documents will only be print for document in that language'),
    }
insurance_contract_type_specific_conditions_attachments()


class insurance_contract(osv.osv):
    _name = 'insurance.contract'
    _rec_name = 'reference'

    _states_selection = [
        ('draft', 'Draft'),
        ('pending_validation', 'Pending Validation'),
        ('pending_2nd_validation', 'Pending 2nd Validation'),
        ('validated', 'Validated'),
        ('refused', 'Refused'),
        ('deprecated', 'Deprecated'),
        ('cancelled', 'Cancelled'),
        ('renewed', 'Renewed'),
    ]

    _coverage_status_selection = [
       ('draft', 'Draft'),
       ('waiting_payment', 'Waiting Payment'),
       ('insured', 'Insured'),
       ('none', 'None'),
    ]

    _coverage_status_conditions = {
        'draft': [('state', 'in', ['draft', 'pending_validation', 'pending_2nd_validation'])],
        'waiting_payment': [('state', '=', 'validated'),('paid','=',False)],
        'insured': [('state', '=', 'validated'),('paid','=',True)],
        'none': [('state', 'in', ['refused', 'cancelled', 'renewed', 'deprecated'])],
    }

    def _get_coverage_status(self, cr, uid, ids, fname, args, context=None):
        def cond_eval(obj, left, op, right):
            optest = {
                '=': lambda a, b: operator.eq(a, b),
                'in': lambda a, b: operator.contains(b, a),
            }
            left_value = getattr(obj, left)
            return optest[op](left_value, right)
        res = dict.fromkeys(ids, '')
        for contract in self.browse(cr, uid, ids, context=context):
            for state in ['draft', 'waiting_payment', 'insured', 'none']:
                if all((cond_eval(contract, *x) for x in self._coverage_status_conditions[state] )):
                    res[contract.id] = state
                    break
        return res
        pass

    def _search_coverage_status(self, cr, uid, ids, fname, args, context=None):
        for left, op, right in args:
            if left == 'coverage_status':
                if op == '=':
                    return self._coverage_status_conditions[right]
                if op == 'in':
                    return self._coverage_status_conditions[right[0]]
        return []

    def _get_contract_price(self, cr, uid, ids, fname, args, context=None):
        res = dict.fromkeys(ids, 0.0)
        return res

    def _get_amount_all_from_record(self, cr, uid, contract_lines, context=None):
        tax_pool = self.pool.get('account.tax')
        res = {
            'amount_untaxed': 0.0,
            'amount_tax': 0.0,
            'amount_total': 0.0,
        }
        for l in contract_lines:
            res['amount_untaxed'] += l.price_subtotal
            tres = tax_pool.compute_all(cr, uid, l.tax_ids, l.price_unit, l.product_qty)
            res['amount_tax'] += tres.get('total_included', 0.0) - tres.get('total', 0.0)
        res['amount_total'] = res['amount_untaxed'] + res['amount_tax']
        res['price'] = res['amount_total']
        return res

    def _get_amount_all(self, cr, uid, ids, fname, args, context=None):
        res = {}
        tax_pool = self.pool.get('account.tax')
        for contract in self.browse(cr, uid, ids, context=context):
            res[contract.id] = self._get_amount_all_from_record(cr, uid, contract.line_ids, context=context)
        return res

    def _is_contract_paid(self, cr, uid, ids, fname, args, context=None):
        res = dict.fromkeys(ids, False)
        cr.execute("SELECT id, paid FROM %s WHERE id IN %%s" % (self._table), (tuple(ids),))
        for r in cr.dictfetchall():
            res[r['id']] = r['paid']
        return res

    def _set_contract_paid(self, cr, uid, id, fname, fvalue, args, context=None):
        cr.execute("UPDATE %s SET paid = %s WHERE id = %s" % (self._table, bool(fvalue), id))
        return True

    def onchange_policyholder_id(self, cr, uid, ids, policyholder_id, context=None):
        policyholder_address_id = False
        if policyholder_id:
            addr_result = self.pool.get('res.partner').address_get(cr, uid, [policyholder_id], adr_pref=['invoice'])
            if addr_result and addr_result.get('invoice'):
                policyholder_address_id = addr_result.get('invoice')

        return {'value': {'policyholder_address_id': policyholder_address_id}}

    def onchange_contract_start_date(self, cr, uid, ids, type_id, start_date=None, context=None):
        ocv = {'value': {}}
        if not type_id:
            return ocv
        type_record = self.pool.get('insurance.contract.type').browse(cr, uid, type_id, context=context)
        # check start_date
        if not start_date:
            default_values = self.default_get(cr, uid, ['start_date'], context=context)
            if default_values.get('start_date'):
                start_date = default_values['start_date']
            else:
                start_date = time.strftime(DEFAULT_SERVER_DATE_FORMAT)
            ocv['value']['start_date'] = start_date
        if type_record.forced_start_date:
            start_date = type_record.forced_start_date
            ocv['value']['start_date'] = start_date
        # update end_date
        result_end_date = type_record.get_end_date_str(start_date=start_date)
        ocv['value']['end_date'] = result_end_date.get(type_record.id, False)
        return ocv

    def onchange_contract_type(self, cr, uid, ids, type_id, policyholder_id, reinsurer_id, start_date, end_date, create_date=False, context=None):
        """
        Set some default value on contract when changing type

        :param cr: database cursor
        :param uid: current user id
        :param ids: id or list of ids on which the onchange are working (only the 1st one is considered)
        :param type_id: the new contract type
        :param policyholder_id: the current policyholder
        :param reinsurer_id: the current reinsurer_id (forced if empty or type specify that reinsurer is needed)
        :param start_date: the current start_date (forced even if already filled)
        :param end_date: the current end_date (forced even if already filled)
        """
        ocv = {'value': {}}
        if not type_id:
            return ocv
        contract_line_proxy = self.pool.get('insurance.contract.line')
        type_record = self.pool.get('insurance.contract.type').browse(cr, uid, type_id, context=context)
        now = datetime.now()
        # force to update start_date + end_date
        dates_ocv = self.onchange_contract_start_date(cr, uid, [], type_id, start_date=None, context=context)
        ocv['value'].update(dates_ocv.get('value',{}))
        # update reinsurer_id
        if not reinsurer_id or not create_date:
            # current value is empty or contract hasn't been created
            ocv['value']['reinsurer_id'] = type_record.reinsurer_id.id
        # update hidden attr fields
        ocv['value']['validation_type'] = type_record.validation_type
        ocv['value']['reinsurer_needed'] = type_record.reinsurer_needed
        if not create_date:
            if type_record.tmpl_line_ids:
                tmpl_line_data = []
                for line in type_record.tmpl_line_ids:
                    ocv2 = contract_line_proxy.onchange_product_id(cr, uid, [], line.product_id.id, None, policyholder_id, context=context)
                    line_vals = ocv2.get('value', {})
                    line_vals.update({
                        'product_uom': line.product_uom.id,
                        'price_unit': line.price_unit,
                        'account_id': line.account_id.id,
                        'tax_ids': [ tax.id for tax in line.tax_ids ],
                    })
                    ocv2 = contract_line_proxy.onchange_product_qty(cr, uid, [], line.product_id.id, line_vals['product_uom'], line.product_qty, line_vals['price_unit'], context=context)
                    line_vals.update(ocv2.get('value', {}))
                    line_vals.update({
                        'product_id': line.product_id.id,
                        'product_qty': line.product_qty,
                    })
                    tmpl_line_data.append(line_vals)
                ocv['value']['line_ids'] = tmpl_line_data
            else:
                ocv['value']['line_ids'] = []
            amount_dict = self._get_amount_all_from_record(cr, uid, type_record.tmpl_line_ids, context=context)
            if amount_dict:
                ocv['value'].update(amount_dict)

        ocv['value']['deductible_amount'] = type_record.deductible_amount
        ocv['value']['clause_ids'] = [ clause.id for clause in type_record.tmpl_clause_ids ]
        return ocv

    def copy_data(self, cr, uid, id, default=None, context=None):
        orm_copied_vals = super(insurance_contract, self).copy_data(cr, uid, id, default=default, context=context)
        orm_copied_vals['paid'] = False # force: non paid

        fields_to_remove_from_copy = [
            'claim_ids',
            'expert_report_ids',
            'invoice_id',
            'version',
            'validate_date',
            'refusal_date',
            'refusal_reason',
            'cancel_date',
            'cancel_reason',

        ]
        for f in fields_to_remove_from_copy:
            orm_copied_vals.pop(f, None)

        return orm_copied_vals

    def name_get(self, cr, user, ids, context=None):
        if not context:
            context = {}
        if not ids:
            return []
        if isinstance(ids, (int, long)):
            ids = [ids]
        return [(r['id'], tools.ustr(r[self._rec_name] or '%s*' % (r['id']))) for r in self.read(cr, user, ids,
            [self._rec_name], context, load='_classic_write')]

    def name_search(self, cr, user, name='', args=None, operator='ilike', context=None, limit=100):
        """
        Search for records and their display names according to a search domain.
        """
        if args is None:
            args = []
        if context is None:
            context = {}
        args = args[:]
        if name:
            if name.isdigit() and -2147483648 <= int(name) <= +2147483647:
                args += ['|',(self._rec_name, operator, name),('id', '=', name)]
        ids = self._search(cr, user, args, limit=limit, context=context)
        res = self.name_get(cr, user, ids, context)
        return res

    _columns = {
        'reference': fields.char('Reference', size=32, readonly=True),
        'version': fields.integer('Version', readonly=True),
        'type_id': fields.many2one('insurance.contract.type', 'Type', required=True),
        'state': fields.selection(_states_selection, 'State', required=True, readonly=True, size=32),
        'coverage_status': fields.function(_get_coverage_status, type='selection', string='Coverage Status',
                                        selection=_coverage_status_selection,
                                        fnct_search=_search_coverage_status, method=True, readonly=True),
        # Workflow related parameters
        'validation_type': fields.related('type_id', 'validation_type', type='selection',
                                        selection=insurance_contract_type._validation_types_selection, readonly=True),
        'insured_amount': fields.float('Insured Amount', digits_compute=dp.get_precision('Amount')),
        'policy_number': fields.char('Policy Number', size=64),
        # Policyholder 
        'policyholder_id': fields.many2one('res.partner', 'Policyholder', required=True),
        'policyholder_address_id': fields.many2one('res.partner.address', 'Policyholder Address', required=True),
        # Reinsurance
        'reinsurer_id': fields.many2one('res.users', 'Reinsurer'),
        'reinsurer_remarks': fields.text('Reinsurer Remarks', help='Only the reinsurer and insurance employee can see that field'),
        'reinsurer_needed': fields.related('type_id', 'reinsurer_needed', type='boolean', string='Reinsurer Needed', readonly=True),
        # Broker
        'broker_id': fields.many2one('res.users', 'Broker'),
        'broker_remarks': fields.text('Broken Remarks', help='Only the broker and insurance employee can see that field'),
        # States log
        'create_date': fields.datetime('Creation Date', readonly=True),
        'validate_date': fields.datetime('Validation Date', readonly=True),
        'refusal_date': fields.datetime('Refusal Date', readonly=True),
        'refusal_reason': fields.text('Refusal Reason', readonly=True),
        'cancel_date': fields.datetime('Cancellation Date', readonly=True),
        'cancel_reason': fields.text('Cancellation Reason', readonly=True),
        # Contract information
        'description': fields.text('Description'),
        'internal_remarks': fields.text('Internal Remarks', read=['base.group_user'], write=['base.group_user']),
        'start_date': fields.date('Start Date',
                                help='Date from where the contract take effect (this correspond to the maturity date of the contract)'),
        'end_date': fields.date('End Date',
                                help='This field indicated the expiry date of the contract'),
        'paid': config_handled_fields_function(_is_contract_paid, type='boolean', string='Paid', fnct_inv=_set_contract_paid, method=True, store=True),
        # TODO: expert reports (one2many ...)
        'expert_report_ids': fields.one2many('insurance.expert.report', 'contract_id', 'Expert Reports', readonly=True),
        'claim_ids': fields.one2many('insurance.claim', 'contract_id', 'Claims'),
        # Invoicing
        'deductible_amount': fields.float('Deductible Amount', digits_compute=dp.get_precision('Amount')),
        'line_ids': fields.one2many('insurance.contract.line', 'contract_id', 'Line'),
        'clause_ids': fields.many2many('insurance.contract.clause', 'contract_id', 'contract_clause_id', 'insurance_contract_clause_rel', 'Clauses'),
        'price': fields.function(_get_amount_all, type='float', string='Price', method=True, multi='amount', store=True, digits_compute=dp.get_precision('Amount')),
        'amount_untaxed': fields.function(_get_amount_all, type='float', string='Amount Untaxed', method=True, multi='amount', digits_compute=dp.get_precision('Amount')),
        'amount_tax': fields.function(_get_amount_all, type='float', string='Amount Tax', method=True, multi='amount', digits_compute=dp.get_precision('Amount')),
        'amount_total': fields.function(_get_amount_all, type='float', string='Amount Total', method=True, multi='amount', digits_compute=dp.get_precision('Amount')),
        'invoice_id': fields.many2one('account.invoice', 'Invoice', readonly=True),
        'specific_conditions': fields.text('Specific Conditions', help='Allow to specify additional conditions to be printed on the contrat'),
    }

    _defaults = {
        'state': 'draft',
        'version': 1,
        'start_date': lambda *a: time.strftime(DEFAULT_SERVER_DATE_FORMAT),
    }

    def _action_draft(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'draft'}, context=context)
    def _action_pending_validation(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'pending_validation'}, context=context)

    def _action_pending_2nd_validation(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'pending_2nd_validation'}, context=context)

    def _action_validate(self, cr, uid, ids, context=None):
        contract_number = self.pool.get('ir.sequence').get(cr, uid, 'insurance.contract')
        contract_vals = {
            'reference': contract_number,
            'validate_date': time.strftime(DEFAULT_SERVER_DATETIME_FORMAT),
            'state': 'validated',
        }
        self.write(cr, uid, ids, contract_vals, context=context)

    def _action_refuse(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'refused'}, context=context)
    def _action_deprecate(self,  cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'deprecated'}, context=context)
    def _action_renew(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'renewed'}, context=context)
    def _action_cancel(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'cancelled'})

    def button_reset_to_draft(self, cr, uid, ids, context=None):
        workflow = netsvc.LocalService('workflow')
        for oid in ids:
            workflow.trg_create(uid, self._name, oid, cr)

    def button_renew(self, cr, uid, ids, context=None):
        # TODO:
        return {
            'type': 'ir.actions.act_window',
            'res_model': 'insurance.contract.cancel.reason.wizard',
            'view_type': 'form',
            'view_mode': 'form',
            'target': 'new',
        }

    def button_create_invoice(self, cr, uid, ids, context=None):
        return False

insurance_contract()

class insurance_contract_line(osv.osv):
    _name = 'insurance.contract.line'

    def _get_price_subtotal(self, cr, uid, ids, fname, args, context=None):
        res = dict.fromkeys(ids, 0.0)
        for l in self.browse(cr, uid, ids, context=context):
            res[l.id] = l.product_qty * l.price_unit
        return res

    def onchange_product_id(self, cr, uid, ids, product_id, product_uom, policyholder_id=None, context=None):
        if context is None:
            context = {}
        if policyholder_id:
            policyholder = self.pool.get('res.partner').browse(cr, uid, policyholder_id, context=context)
            if policyholder.lang:
                context['lang'] = policyholder.lang
        property_pool = self.pool.get('ir.property')
        ocv = { 'value': {}}
        if not product_id:
            ocv['value'].update({
                'product_uom': False,
                'account_id': False,
                'tax_ids': False,
            })
            return ocv
        p = self.pool.get('product.product').browse(cr, uid, product_id, context=context)
        ocv['value'].update({
            'price_unit': p.list_price,
            'product_uom': p.uom_id.id or product_uom or False,
            'description': p.name,
            'note': p.description,
            'tax_ids': [ t.id for t in p.taxes_id ],
            'account_id': p.product_tmpl_id.property_account_income.id or p.categ_id.property_account_income_categ.id or False,
        })
        return ocv

    def onchange_product_qty(self, cr, uid, ids, product_id, product_uom, product_qty, price_unit, context=None):
        ocv = {'value': {}}
        # TODO: take care of product_id, product_uom
        ocv['value']['price_subtotal'] = product_qty * price_unit
        return ocv

    _columns = {
        'contract_id': fields.many2one('insurance.contract', 'Contract', required=True, ondelete='cascade'),
        'product_id': fields.many2one('product.product', 'Product', required=True),
        'account_id': fields.many2one('account.account', 'Account', required=True),
        'tax_ids': fields.many2many('account.tax', 'insurance_contract_line_tax_rel', 'contract_line_id', 'tax_id', 'Taxes', domain=[('parent_id','=',False),('type_tax_use','in',['sale','all'])]),
        'product_uom': fields.many2one('product.uom', 'UoM'),
        'product_qty': fields.float('Quantity', required=True),
        'description': fields.char('Description', size=128, required=True),
        'price_unit': fields.float('Unit Price', required=True, digits_compute=dp.get_precision('Account')),
        'price_subtotal': fields.function(_get_price_subtotal, type='float', method=True, string='Subtotal', digits_compute=dp.get_precision('Account')),
        'note': fields.text('Note'),
    }

    _defaults = {
        'product_qty': 1.0,
    }
insurance_contract_line()

class insurance_contract_ask_reason_wizard(osv.osv):
    _reason_field = ''
    _reason_wkf_signal = ''
    _columns = {
        'reason': fields.text('Reason'),
    }

    def action_confirm(self, cr, uid, ids, context=None):
        close_action = { 'type': 'ir.actions.act_window.close' }
        if not context:
            return close_action
        active_id = context.get('active_id')
        if active_id is None:
            return close_action
        wizard = self.browse(cr, uid, ids[0], context=context)
        f = self._reason_field
        contract_data = {
            f+'_reason': wizard.reason,
            f+'_date': time.strftime(DEFAULT_SERVER_DATETIME_FORMAT),
        }
        contract_proxy = self.pool.get('insurance.contract')
        contract_proxy.write(cr, uid, [active_id], contract_data)
        if self._reason_wkf_signal:
            workflow = netsvc.LocalService('workflow')
            workflow.trg_validate(uid, 'insurance.contract', active_id, self._reason_wkf_signal, cr)
        return close_action

class insurance_contract_refuse_reason_wizard(insurance_contract_ask_reason_wizard):
    _name = 'insurance.contract.refuse.reason.wizard'
    _reason_field = 'refusal'
    _reason_wkf_signal = 'signal_refuse'
insurance_contract_refuse_reason_wizard()

class insurance_contract_cancel_reason_wizard(insurance_contract_ask_reason_wizard):
    _name = 'insurance.contract.cancel.reason.wizard'
    _reason_field = 'cancel'
    _reason_wkf_signal = 'signal_cancel'
insurance_contract_cancel_reason_wizard()

class insurance_expert_report_type(osv.osv):
    _name = 'insurance.expert.report.type'
    _types_selection = [
        ('on_contract', 'On Contract'),
        ('on_claim', 'On Claim'),
    ]

    _columns = {
        'name': fields.char('Name', size=64, required=True, translate=True,),
        'code': fields.char('Code', size=32, required=True),
        'description': fields.text('Description'),
        'type': fields.selection(_types_selection, 'Type', required=True),
        'sequence_id': fields.many2one('ir.sequence', 'Sequence', domain=[('code','=','insurance_expert_report')]),
        'has_limited_validity': fields.boolean('Has limited validity',
                                            help='If check, for expert report using this type, the date start and date stop fields will be shown'),
    }

    _defaults = {
        'type': 'on_contract',
    }

    _sql_constraints = [
        ('uniq_code', 'UNIQUE(code)', 'Code must be unique'),
    ]
insurance_expert_report_type()

class insurance_claim(osv.osv):
    _name = 'insurance.claim'
    _rec_name = 'date'

    _states_selection = [
        ('draft', 'Draft'),
        ('inprogress', 'In Progress'),
        ('received', 'Received'),
        ('accepted', 'Accepted'),
        ('refused', 'Refused'),
    ]

    _columns = {
        'create_date': fields.datetime('Creation Date', readonly=True),
        'contract_id': fields.many2one('insurance.contract', 'Contract', required=True),
        'date': fields.datetime('Claim Date', required=True),
        'expert_note': fields.text('Expert Note'),
        'broker_note': fields.text('Broker Note'),
        'reinsurer_note': fields.text('Reinsurer Note'),
        'internal_note': fields.text('Internal Note', read=['base.group_user'], write=['base.group_user']),
        'expert_report_ids': fields.one2many('insurance.expert.report', 'claim_id', 'Expert Reports'),
        'state': fields.selection(_states_selection, 'State', readonly=True),
    }

    _defaults = {
        'state': 'draft',
    }

    def _action_draft(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'draft'}, context=context)

    def _action_inprogress(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'inprogress'})

    def _action_receive(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'received'}, context=None)

    def _action_accept(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'accepted'}, context=context)

    def _action_refuse(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'refused'}, context=context)

insurance_claim()

class insurance_expert_report(osv.osv):
    _name = 'insurance.expert.report'
    _rec_name = 'id'
    _states_selection = [
        ('draft', 'Draft'),
        ('validated', 'Validated'),
        ('deprecated', 'Deprecated'),
        ('cancelled', 'Cancelled'),
    ]
    _columns = {
        'id': fields.integer('Id', readonly=True),
        'type_id': fields.many2one('insurance.expert.report.type', 'Type', required=True),
        'type_mode': fields.related('type_id', 'type', type='selection', selection=insurance_expert_report_type._types_selection, string='Type Mode', readonly=True),
        'type_has_limited_validity': fields.related('type_id', 'has_limited_validity', type='boolean', string='Type has limited validity', readonly=True),
        'claim_id': fields.many2one('insurance.claim', 'Claim'),
        'contract_id': fields.many2one('insurance.contract', 'Contract'),
        'policyholder_id': fields.related('contract_id', 'policyholder_id', type='many2one', relation='res.partner', string='Policyholder', readonly=True),
        'responsible_id': fields.many2one('res.users', 'Responsible', required=True),
        'responsible_summary': fields.text('Responsible Summary'),
        'create_date': fields.datetime('Creation Date', readonly=True),
        'validate_date': fields.datetime('Validation Date', readonly=True),
        'start_date': fields.date('Start Date'),
        'end_date': fields.date('End Date'),
        'internal_note': fields.text('Internal Note', read=['base.group_user'], write=['base.group_user']),
        'state': fields.selection(_states_selection, 'State', readonly=True),
    }

    _defaults = {
        'state': 'draft',
        'responsible_id': lambda self, cr, uid, context: uid,
    }

    def onchange_report_type(self, cr, uid, ids, type_id, context=None):
        """
        update expert report values when user change the expert report type

        :param cr: database cursor
        :param uid: the current user id
        :param ids: the id or list of ids the onchange callback is working on
        :param type_id: the new type_id
        :param context: the current context of execution
        """
        ocv = {'value': {}}
        if not type_id:
            return ocv
        type_record = self.pool.get('insurance.expert.report.type').browse(cr, uid, type_id, context=context)
        ocv['value'].update({
            'type_mode': type_record.type,
            'type_has_limited_validity': type_record.has_limited_validity,
        })
        return ocv

    def onchange_report_claim(self, cr, uid, ids, type_id, claim_id, context=None):
        """
        update expert report contract_id when user change the expert report claim

        :param cr: database cursor
        :param uid: the current user id
        :param ids: the id or list of ids the onchange callback is working on
        :param type_id: the current expert report type id
        :param claim_id: the new claim_id
        :param context: the current context of execution
        """
        ocv = {'value': {}}
        if type_id:
            type_record = self.pool.get('insurance.expert.report.type').browse(cr, uid, type_id, context=context)

        if not claim_id:
            if type_id and not type_record.type == 'on_contract':
                # force empty contract_id when removing claim field
                # and expert report type does not enforce contract_id field
                ocv['value']['contract_id'] = False
            return ocv

        claim_record = self.pool.get('insurance.claim').browse(cr, uid, claim_id, context=context)
        if type_id and type_record.type == 'on_contract':
            ocv['value']['contract_id'] = claim_record.contract_id.id
        return ocv

    def _action_draft(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'draft'}, context=context)

    def _action_validate(self, cr, uid, ids, context=None):
        expert_report_vals = {
            'state': 'validated',
            'validate_date': time.strftime(DEFAULT_SERVER_DATETIME_FORMAT),
        }
        self.write(cr, uid, ids, expert_report_vals, context=context)

    def _action_cancel(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'cancelled'}, context=context)

    def _action_deprecate(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'deprecated'}, context=context)

insurance_expert_report()
